use super::state::State;
use std::{sync::atomic::Ordering, thread};

fn build_led_package(len: u32, color: (u8, u8, u8)) -> Vec<u8> {
    let mut packet: Vec<u8> = vec![1, 1];

    // loop through leds
    for i in 0..len {
        packet.push(i as u8);
        packet.push(color.0);
        packet.push(color.1);
        packet.push(color.2);
    }

    return packet;
}

const FLASHED_X2: f64 = 0.00003113;
const FLASHED_X1: f64 = -0.17396053;
const FLASHED_X0: f64 = 244.51964267;

const BOMB_MAX_TIME: u64 = 40_000;
const BOMB_BLINK_DURATION: u64 = 300;
const BOMB_BLINK_SPEED: u64 = 700;

fn interpolate_flashed(time: u64) -> u8 {
    if time == 0 {
        return 0;
    }

    let y = FLASHED_X2 * (time as f64).powi(2) + FLASHED_X1 * (time as f64) + FLASHED_X0;

    return y.min(255.0).round() as u8;
}

fn resolve_time_from_flashed(flashed: u8) -> u64 {
    let discriminant = FLASHED_X1.powi(2) - 4.0 * FLASHED_X2 * (FLASHED_X0 - flashed as f64);

    let x = (-FLASHED_X1 - discriminant.sqrt()) / (2.0 * FLASHED_X2);
    return x.min(5000.0) as u64;
}

pub fn led_loop(fps: u32, addr: &str, len: u32, state: State) {
    let socket = std::net::UdpSocket::bind("0.0.0.0:0").unwrap();
    let sleep_time: u64 = (1000 / fps).into();

    let mut last_up_flashed: u8 = 0; // Used to detect changes in upstream flashed
    let mut interpolated_flashed: u8;
    let mut time_since_last_flashed: u64 = 0;

    let mut last_up_planted: bool = false;
    let mut time_bomb_planted: u64 = 0;

    loop {
        thread::sleep(std::time::Duration::from_millis(sleep_time));

        let upstream_flashed = u8::try_from(state.flashed.load(Ordering::Relaxed)).unwrap_or(0);

        // Count the time since the flash started to fade
        if upstream_flashed > 0 && upstream_flashed < 255 {
            time_since_last_flashed += sleep_time;
        }

        if upstream_flashed == last_up_flashed {
            // Upstream value hasn't changed

            if upstream_flashed == 255 {
                // Flash is at max
                interpolated_flashed = 255;
            } else if upstream_flashed == 0 {
                // Flash is off
                interpolated_flashed = 0;
            } else {
                // Flash is fading
                interpolated_flashed = interpolate_flashed(time_since_last_flashed);
            }
        } else {
            // Upstream value has changed

            time_since_last_flashed = resolve_time_from_flashed(upstream_flashed);
            interpolated_flashed = upstream_flashed;
            last_up_flashed = upstream_flashed;
        }

        let pack = build_led_package(
            len,
            (
                interpolated_flashed,
                interpolated_flashed,
                interpolated_flashed,
            ),
        );
        socket.send_to(&pack, addr).unwrap();

        // let planted = state.bomb_planted.load(Ordering::Relaxed);
        let planted = true;


        if planted != last_up_planted {
            // A bomb has been planted or defused reset the timer
            time_bomb_planted = 0;
            last_up_planted = planted;
        }

        if planted {
            time_bomb_planted += sleep_time;

            println!("Time: {}", time_bomb_planted);
            
            let should_blink = (time_bomb_planted) % (BOMB_BLINK_SPEED * (BOMB_MAX_TIME / time_bomb_planted).max(1))  < BOMB_BLINK_DURATION;

            let color = if should_blink {
                (255, 0, 0)
            } else {
                (0, 0, 0)
            };

            let pack = build_led_package(len, color);
            socket.send_to(&pack, addr).unwrap();
    
        }
    }
}
